import logging
import re
import time
from collections import namedtuple
from typing import List, NamedTuple, Union

import requests

from ...client import Session
from ...models import Block, BlockFormat, BlockVersion
from .regex import INLINE_CITATION_BLOCK_REGEX

logger = logging.getLogger()

LocalReferenceItem = namedtuple(
    "LocalReferenceItem", ["block_path", "local_tag", "bibtex"]
)


class LocalMarkerItem(NamedTuple):
    remote_marker: str
    local_marker: str


def find_groups_in_content(regex_matcher: str, content: str):
    """
    Take a peice of tex content that we expect to contain one or more commands,
    find all instances based on the regex given and return the first groups

    Typically this is used to extract block_paths for different latex commands from
    TeX content
    """
    block_paths = []
    matches = re.finditer(regex_matcher, content)
    for match in matches:
        block_path = match[1]
        block_paths.append(block_path)

    return block_paths


def block_hash_to_url(
    api_url: str, block_hash: str, format: Union[BlockFormat, None] = None
):
    fmt = ""
    if format is not None:
        fmt = f"?format={format}"
    match = re.search(
        r"block:([A-Za-z0-9]{20})/([A-Za-z0-9]{20})/*([0-9]*)", block_hash
    )
    if match:
        if match[3]:
            return f"{api_url}/blocks/{match[1]}/{match[2]}/versions/{match[3]}{fmt}"
        return f"{api_url}/blocks/{match[1]}/{match[2]}{fmt}"
    raise ValueError(f"invalid block hash {block_hash}")


# TODO move to session - easier to mock
def get_model(session, url, model=BlockVersion):
    block = session._get_model(url, model)
    if not block:
        raise ValueError(f"Could not fetch the block {url}")
    return block


def fetch(url: str):
    resp = requests.get(url)
    if resp.status_code >= 400:
        raise ValueError(resp.content)
    return resp.content


def localize_references_from_content_block(
    session: Session, reference_list: List[LocalReferenceItem], content: str
):
    """Looks for cite TeX commands in the content then replaces the block ids
    with locally unique identifiers based on the local reference list.

    The reference list is extended as new references are found (side effect)

    Appends a unique hash to each new reference encountered
    """
    block_paths = find_groups_in_content(INLINE_CITATION_BLOCK_REGEX, content)
    patched_content = content

    for block_path in block_paths:
        # check for the reference in the reference list based on the block_path
        matched_references = [r for r in reference_list if r.block_path == block_path]
        existing_reference = (
            matched_references[0] if (len(matched_references) > 0) else None
        )

        if existing_reference is None:
            url = block_hash_to_url(session.api_url, block_path)
            logging.info(f"fetching reference block {url}")
            block = get_model(session, url, Block)
            logging.info("got reference block")

            # get latest version
            version_url = f"{url}/versions/{block.latest_version}?format=bibtex"
            logging.info(f"fetching reference version {version_url}")
            version = get_model(session, version_url)
            logging.info("got reference version")

            # update the list
            local_tag, plain_tag = parse_cite_tag_from_version(version.content)
            bibtex = version.content.replace(plain_tag, local_tag)

            reference_item = LocalReferenceItem(block_path, local_tag, bibtex)
            reference_list.append(reference_item)
            existing_reference = reference_item
            logging.info(f"using new reference {existing_reference.local_tag}")
        else:
            logging.info(f"using existing reference {existing_reference.local_tag}")

        # patch the content and move on
        patched_content = patched_content.replace(
            block_path, existing_reference.local_tag
        )

    return patched_content


def get_fast_hash():
    return hex(int(time.time_ns()))[2:]


def parse_cite_tag_from_version(content: str):
    tag = "ref"
    match = re.match("@article{([0-9a-zA-Z_]+),*\n", content)
    if match is not None:
        tag = match[1].replace(",", "")
    # adding quasi-random hash to save de-duping work
    hash_id = get_fast_hash()
    tag_with_hash = f"{tag}_{hash_id}"

    return tag_with_hash, tag



from ...client import Session
from .regex import INLINE_IMAGE_BLOCK_REGEX
from .patch_local_images_from_tex_block import patch_local_images_from_tex_block
from .get_image_block import get_image_block


def localize_images_from_content_block(
    session: Session, assets_folder: str, content: str
):
    """Process image blocks and their respective referencing data including:
    - captions
    - labels
    """
    patch = patch_local_images_from_tex_block(INLINE_IMAGE_BLOCK_REGEX, content)
    content_with_extensions = patch.content

    for block_path, local_path in zip(patch.block_paths, patch.local_paths):
        # get block
        image_block, local_path_with_extension = get_image_block(
            session, assets_folder, block_path, local_path
        )

        # now patch the image captions which will be marked have placeholders like
        # f"{local_path}.caption"
        content_with_extensions = content_with_extensions.replace(
            f"{local_path}.caption", image_block.caption
        )

        # patch the remaining image block references
        content_with_extensions = content_with_extensions.replace(
            local_path, local_path_with_extension
        )

    return content_with_extensions



import re
from typing import List
from .localize_images import ImageSummary


def patch_local_images_from_tex_block(regex_matcher: str, content: str):
    """
    Take a peice of tex content that we expect to contain only one includegraphics directive
    get the remote path, assign a new local path and retrun a structure that can be parsed
    """
    updated_content = content
    block_paths: List[str] = []
    local_paths: List[str] = []
    matches = re.finditer(regex_matcher, content)
    for match in matches:
        block_path = match[1]
        hash_segment = block_path[6:]
        local_segment = hash_segment.replace("/", "_")
        local_path = block_path.replace("block:", "images/").replace(
            hash_segment, local_segment
        )
        updated_content = updated_content.replace(block_path, local_path)
        block_paths.append(block_path)
        local_paths.append(local_path)

    return ImageSummary(
        content=updated_content,
        block_paths=block_paths,
        local_paths=local_paths,
    )



from curvenote.models import (
    ArticleVersionPost,
    BlockChild,
    BlockFormat,
    BlockKind,
    BlockPost,
    ContentVersionPost,
)
import logging
import os
import sys
import subprocess
import typer

from .version import __version__
from .client import Session
from .latex.project import LatexProject

app = typer.Typer()

logger = logging.getLogger()


def version_callback(value: bool):
    if value:
        typer.echo(
            r"""
   ______                                  __
  / ____/_  ________   _____  ____  ____  / /____
 / /   / / / / ___/ | / / _ \/ __ \/ __ \/ __/ _ \
/ /___/ /_/ / /   | |/ /  __/ / / / /_/ / /_/  __/
\____/\__,_/_/    |___/\___/_/ /_/\____/\__/\___/
        """
        )
        typer.echo(f"Curvenote CLI Version: {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    )
):
    return


@app.command()
def get_me(
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    )
):
    session = Session(token)
    typer.echo(session.user().json(indent=4))


@app.command()
def get_my_projects(
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    )
):
    session = Session(token)
    for project in session.my_projects():
        typer.echo(project.json(indent=4))


@app.command()
def push(
    path: str = typer.Argument(..., help="Local file or folder to push to curvenote."),
    project: str = typer.Argument(
        ...,
        help=(
            "Identifier of target Project. PROJECT may match title, name, or id "
            "of an existing Project. If no existing Project is found, a new "
            "Project will be created with title PROJECT."
        ),
    ),
    article: str = typer.Option(
        None,
        help=(
            "Identifier of target Article. ARTICLE may match title, name, or id "
            "of an existing Article. If no existing Article is found, a new "
            "Article will be created with title ARTICLE. ARTICLE is ignored if "
            "PATH is a folder. If PATH is a folder or ARTICLE is not provided, "
            "filename will be used for Article."
        ),
    ),
    team: str = typer.Option(
        None,
        help=(
            "Team to use when creating a new Project. TEAM is ignored if PROJECT "
            "already exists. If PROJECT does not exist and TEAM is not provided, "
            "the new Project will be created under the current user."
        ),
    ),
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    ),
):
    """Push contents of local file or folder to curvenote Project"""
    if not os.path.exists(path):
        raise ValueError(f"path not found: {path}")
    session = Session(token)

    typer.echo("Checking for project access")
    project_obj = session.get_or_create_project(
        title=project,
        team=team,
    )
    if os.path.isdir(path):
        typer.echo("pushing from folder")
        session.documents_from_folder(folder=path, project=project_obj)
    elif os.path.isfile(path):
        _, file_extension = os.path.splitext(path)
        if file_extension == ".ipynb":
            typer.echo("pushing notebook file...")
            session.notebook_from_file(
                filename=path, project=project_obj, title=article
            )
        elif file_extension == ".md":
            typer.echo("pushing markdown file...")
            session.article_from_file(filename=path, project=project_obj, title=article)
        else:
            raise ValueError(f"unsupported file type: {file_extension}")
    else:
        raise ValueError(f"unable to resolve path: {path}")


@app.command()
def pull_as_latex(
    target: str = typer.Argument(
        ...,
        help=(
            "Local folder in which to construct the Latex assets. If TARGET exists it"
            "and all files will be removed and a new empty folder structure created"
        ),
    ),
    project: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Project containing ARTICLE. PROJECT may match title,"
            " name, or id of an existing Project. If no existing Project is found, an "
            "error will be raised"
        ),
    ),
    article: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Article. ARTICLE may match title, name, or id "
            "of an existing Article. If no existing Article is found, an error will"
            "be raised."
        ),
    ),
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    ),
    version: int = typer.Option(
        None,
        help=(
            "Version of the article to pull, if not specified will pull the latest version."
        ),
    ),
):
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    session = Session(token)

    try:
        project_obj = session.get_project(project)
    except ValueError as err:
        typer.echo(f"Could not find project: {project} or you do not have access")
        raise typer.Exit(code=1) from err

    typer.echo(f"Found project: {project_obj.name}")

    try:
        LatexProject.build_single_article(
            session, target, project_obj, article, version
        )
    except ValueError as err:
        typer.echo(err)
        raise typer.Exit(code=1)


@app.command()
def build_pdf(
    target: str = typer.Argument(
        ...,
        help=(
            "Local folder containing the local LaTeX project."
            "Must contain an index.tex file."
        ),
    ),
):
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    typer.echo("Build PDF from LaTeX")
    typer.echo(f"Target folder: {target}")
    typer.echo("Invoking xelatex...")
    try:
        XELATEX_CMD=f'cd {target}; latexmk -f -xelatex -synctex=1 -interaction=nonstopmode -file-line-error -latexoption="-shell-escape" index.tex'
        ret_val = subprocess.run(XELATEX_CMD, shell=True)
        if ret_val.returncode != 0:
            typer.echo(f"mklatex returned with a non zero error code, but thePDF was still created")
            typer.Exit(code=1)
    except ValueError as err:
        typer.echo(f"Fatal error while running mklatex")
        typer.Exit(code=1)

    typer.echo(f"mklatex reports success!")
    typer.Exit(code=0)


@app.command()
def clone_content(
    project: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Project containing ARTICLE. PROJECT may match title,"
            " name, or id of an existing Project. If no existing Project is found, an "
            "error will be raised"
        ),
    ),
    article: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Article. ARTICLE may match title, name, or id "
            "of an existing Article. If no existing Article is found, an error will"
            "be raised."
        ),
    ),
    new_article_name: str = typer.Argument(
        ...,
        help=(
            "Title of the new article to be created with cloned content."
            "If an article with the same name is found, an error will be raised."
        ),
    ),
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    ),
    version: int = typer.Option(
        None,
        help=(
            "Version of the article to pull, if not specified will pull the latest version."
        ),
    ),
):
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    session = Session(token)

    typer.echo("Checking for project access")
    try:
        project_obj = session.get_project(project)
    except ValueError as err:
        typer.echo(f"Unable to access project {project}")
        typer.Exit(code=1)

    typer.echo("Fetching existing article block")
    try:
        block = session.get_block(project_obj, article, BlockKind.article)
    except ValueError as err:
        typer.echo(f"Unable to access article {article}")
        typer.Exit(code=1)

    version = int(version) if version is not None else block.latest_version
    typer.echo("Fetching existing article version")
    try:
        version = session.get_version(block, version)
    except ValueError as err:
        typer.echo(f"Unable to get article version {version}")
        typer.echo(str(err))
        typer.Exit(code=1)

    new_order = []
    new_children = {}

    for o in version.order:
        child = version.children[o]
        try:
            child_block = session.get_block(project_obj, child.src.block)
        except ValueError as err:
            typer.echo(f"Unable to access child, skipping {child.src}")
            continue

        if child_block.kind != BlockKind.content:
            typer.echo(f"Found {child_block.kind} block, referencing")
            new_order.append(child.id)
            new_children[child.id] = child
        else:
            typer.echo("Found content block, cloning")
            child_version = session.get_version(child_block, child.src.version)
            new_child_block = session.upload_block(
                BlockPost(kind=child_block.kind), project=project_obj
            )
            new_child_version = session.upload_version(
                version=ContentVersionPost(content=child_version.content),
                block=new_child_block,
            )
            new_order.append(child.id)
            new_children[child.id] = BlockChild(id=child.id, src=new_child_version.id)

    typer.echo(f"Existing article had {len(version.order)} children")
    typer.echo(f"Cloned article will have {len(new_order)} children")

    typer.echo(f"Creating new article")
    try:
        new_article_block = session.upload_block(
            BlockPost(name=new_article_name, kind=BlockKind.article),
            project=project_obj,
        )
    except ValueError as err:
        typer.echo("Could not create new article block")
        typer.echo(str(err))
        typer.Exit(code=1)

    try:
        session.upload_version(
            version=ArticleVersionPost(
                order=new_order,
                children=new_children,
                title=new_article_name,
            ),
            block=new_article_block,
        )
    except ValueError as err:
        typer.echo("Could not create new article version")
        typer.echo(str(err))
        typer.Exit(code=1)

    typer.echo(f"New article created: {new_article_name}")



import os
from typing import Optional

from .models import Block, BlockFormat, Project
from .utils import filename_to_title, title_to_name

PROCESS_NOTEBOOK_URL = "{api_url}/process/notebook"
NOTEBOOK_CONTENT_TYPE = "application/x-ipynb+json"


class Notebook:

    block_format = BlockFormat.jupyter

    def __init__(self, filename: str, session):
        self.session = session
        self.filename = os.path.abspath(filename)

    def upload(self, project: Project, title: Optional[str] = None) -> Block:
        """Upload Jupyter Notebook to curvenote Project

        If ``title`` is not provided, title will be determined from the notebook
        filename.

        Currently this does not support versioning notebook cells or entire
        notebooks; it only uplaods a fresh copy of the notebook.
        """
        if title is None:
            title = filename_to_title(self.filename)
        storage_url = self.session.upload_file(self.filename, NOTEBOOK_CONTENT_TYPE)
        resp = self.session._post(
            PROCESS_NOTEBOOK_URL.format(api_url=self.session.api_url),
            json_dict={
                "project": project.id,
                "name": title_to_name(title),
                "title": title,
                "format": self.block_format,
                "upload_path": storage_url,
            },
        )
        return Block(**resp.json())



import os
import re
from typing import List

from .image import IMAGE_TAG, Image
from .models import (
    ArticleVersionPost,
    Block,
    BlockChild,
    BlockFormat,
    BlockKind,
    BlockPost,
    ContentVersionPost,
    Project,
)
from .utils import github_to_raw, is_url

IMAGE_REGEX = re.compile(r'^!\[(?P<alt>.*)\]\((?P<path>.*?)(\ "(?P<title>.*)")?\)$')


class Markdown:

    block_kind = BlockKind.content
    block_format = BlockFormat.md

    def __init__(self, filename: str, session):
        self.session = session
        self.filename = os.path.abspath(filename)

    def get_sections(self) -> List[str]:
        """Break markdown content into sections to be uploaded as individual blocks

        Image references and code blocks are each given their own section.
        All remaining content is naively split at double line breaks.
        """
        with open(self.filename, "r") as file:
            lines = [line if line.strip() else "" for line in file.read().split("\n")]
        output_sections = []
        section = ""
        tick_code_block = False
        tab_code_block = False
        tex_block = False
        for line in lines:
            if line.strip() == r"\begin{equation}":
                line = "$$"
                tex_block = True
            if line.strip() == r"\end{equation}":
                line = "$$"
                tex_block = False
            if IMAGE_REGEX.match(line):
                if section:
                    output_sections.append(section)
                    section = ""
                output_sections.append(line)
                continue
            if line.startswith("```"):
                tick_code_block = not tick_code_block
            if line.startswith("    ") and not tick_code_block and not tex_block:
                tab_code_block = True
            elif tab_code_block and line:
                tab_code_block = False
                output_sections.append(section.strip("\n"))
                section = ""
            if section and not line and not (tick_code_block or tab_code_block):
                output_sections.append(section)
                section = ""
            elif section:
                section = "\n".join([section, line])
            elif line:
                section = line
        if section:
            output_sections.append(section)
        return output_sections

    def upload(self, article: Block, title: str = None):
        """Upload markdown content to provided article

        Article ``title`` will be set if specified.
        This creates a new version of the article with blocks extracted from the file.
        The extracted blocks are not tied to the blocks in the previous article version.

        Raises ``ValueError`` if upload fails.
        """
        project = self.session.resolve(article.links.project, Project)
        version = ArticleVersionPost(title=title, order=[], children={})
        for i, section in enumerate(self.get_sections()):
            block_format = self.block_format
            image_match = IMAGE_REGEX.match(section)
            if image_match:
                title = image_match.group("title")
                alt = image_match.group("alt")
                path = image_match.group("path")
                if not is_url(path) and not os.path.isabs(path):
                    path = os.path.abspath(
                        os.path.sep.join([os.path.dirname(self.filename), path])
                    )
                try:
                    image = Image(path, self.session)
                except ValueError:
                    if is_url(path):
                        section = IMAGE_TAG.format(
                            title=title or "",
                            alt=alt or "",
                            src=github_to_raw(path),
                            align="center",
                            width=70,
                        )
                    else:
                        raise
                else:
                    section = image.create_tag(project, title, alt)
                section += "<p></p>"
                block_format = BlockFormat.html
            section_block = self.session.upload_block(
                BlockPost(kind=self.block_kind, tags=[]), project
            )
            section_version = self.session.upload_version(
                ContentVersionPost(content=section, format=block_format),
                section_block,
            )
            local_id = str(i)
            version.children.update(
                {local_id: BlockChild(id=local_id, src=section_version.id)}
            )
            version.order.append(local_id)
        return self.session.upload_version(version, article)



from .get_image_block import get_image_block
from ...models import BlockVersion, BlockChildID
from ...client import Session
from .index import get_fast_hash
import logging

VERSION_ID = "___VERSION_ID___"
CAPTION = "___CAPTION___"
LABEL = "___LABEL___"

IMAGE_LATEX_TPL = rf"""\begin{{figure}}[h]
  \centering
  \includegraphics[width=0.4\linewidth]{{{VERSION_ID}}}
  \caption{{{CAPTION}}}
  \label{{{LABEL}}}
\end{{figure}}
"""


def version_id_to_block_path(id: BlockChildID):
    return f"block:{id.project}/{id.block}/{id.version}"


def version_id_to_local_path(id: BlockChildID):
    return f"images/{id.project}_{id.block}_{id.version}"


def localize_image_from_top_level_block(
    session: Session, assets_folder: str, version: BlockVersion
):
    block_path = version_id_to_block_path(version.id)
    local_path = version_id_to_local_path(version.id)

    image_block, local_path_with_extension = get_image_block(
        session, assets_folder, block_path, local_path
    )

    content = (
        IMAGE_LATEX_TPL.replace(VERSION_ID, local_path_with_extension)
        .replace(CAPTION, image_block.caption)
        .replace(LABEL, get_fast_hash())
    )

    return f"\n\n{content}\n"



from curvenote.latex.utils.imagemagick import convert_gif
import logging
import os

import requests

from ...client import Session
from ...models import BlockFormat, BlockVersion
from .index import block_hash_to_url
from .localize_images import ImageFormats


def get_image_block(
    session: Session, assets_folder: str, block_path: str, local_path: str
):
    url = block_hash_to_url(session.api_url, block_path, BlockFormat.tex)
    logging.info("fetching image block from %s", url)
    image_block = session._get_model(url, BlockVersion)
    if not image_block:
        raise ValueError(f"Could not fetch image block {url}")

    # download from links.download (signed) to local path
    if not image_block.links.download:
        raise ValueError(f"Block kind {image_block.kind} has no download link")

    resp = requests.get(image_block.links.download)
    if resp.status_code >= 400:
        raise ValueError(resp.content)

    # now that we have the block we need to update the local path in the content
    # and save the file with the correct file extension
    content_type = resp.headers.get("content-type")
    if content_type not in ImageFormats:
        raise ValueError(f"Unsupported image content type {content_type}")

    ext = ImageFormats[content_type]
    local_path_with_extension = f"{local_path}.{ext}"
    logging.info("Writing %s", local_path_with_extension)
    with open(os.path.join(assets_folder, local_path_with_extension), "wb+") as file:
        file.write(resp.content)

    if content_type == "image/gif":
        logging.info("Found image/gif converting to image/png")
        # convert to png, leave gif file in place but return the png path for use in content
        local_path_with_extension = convert_gif(assets_folder, local_path_with_extension)
        logging.info(f"Updated local path: {local_path_with_extension}")
        # either calls will raise exception on error causing file to be skipped

    return image_block, local_path_with_extension



import logging
from .version import __version__
from functools import lru_cache
from glob import glob
from os import environ, path
from typing import Iterator, List, Optional, Type, Union
from urllib.parse import urlparse, urlunparse

import requests
from pydantic import AnyHttpUrl, BaseModel, ValidationError

from .markdown import Markdown
from .models import (
    Access,
    AccessList,
    ArticleVersionPost,
    Block,
    BlockChild,
    BlockDraft,
    BlockDraftList,
    BlockFormat,
    BlockKind,
    BlockList,
    BlockPost,
    BlockVersion,
    BlockVersionList,
    ChildrenList,
    Comment,
    CommentList,
    ContentVersionPost,
    ImageVersionPost,
    MyUser,
    NavigationItem,
    NavigationVersionPost,
    Project,
    ProjectList,
    ProjectPost,
    User,
)
from .notebook import Notebook
from .utils import filename_to_title, title_to_name, update_query

logger = logging.getLogger()

API_RESPONSE_MODELS = [
    Access,
    AccessList,
    BlockVersion,
    BlockVersionList,
    Block,
    BlockDraft,
    BlockDraftList,
    BlockList,
    Comment,
    CommentList,
    Project,
    ProjectList,
    User,
    MyUser,
]

CLIENT_NAME = "Curvenote Python Client"
CLIENT_VERSION = __version__
if environ.get("CURVENOTE_ENV") == "development":
    print(
        r"""
  ___            __  __         _
 |   \ _____ __ |  \/  |___  __| |___
 | |) / -_) V / | |\/| / _ \/ _` / -_)
 |___/\___|\_/  |_|  |_\___/\__,_\___|

    """
    )
    API_URL = environ.get("CURVENOTE_API", "http://localhost:8083")
    print(f"Accessing API at: {API_URL}")
    SITE_URL = "https://localhost:3000"
else:
    API_URL = environ.get("CURVENOTE_API", "https://api.curvenote.com")
    SITE_URL = "https://curvenote.com"
ME_URL = "{api_url}/my/user"
USERS_URL = "{api_url}/users/{user_id}"
PROJECTS_URL = "{api_url}/projects"
UPLOAD_URL = "{api_url}/my/upload"


class Session:
    def __init__(
        self, token: str, api_url: Optional[str] = None, site_url: Optional[str] = None
    ) -> None:
        self.api_url = api_url or API_URL
        self.site_url = site_url or SITE_URL
        self._session = requests.Session()
        self._session.headers.update(
            {
                "Authorization": f"Bearer {token}",
                "X-Client-Name": CLIENT_NAME,
                "X-Client-Version": CLIENT_VERSION,
            }
        )

    @lru_cache(maxsize=128)
    def _get(self, url: str) -> requests.Response:
        resp = self._session.get(url)
        if resp.status_code >= 400:
            raise ValueError(resp.content)
        return resp

    def _get_model(self, url: str, model: Type[BaseModel]) -> BaseModel:
        resp = self._get(url)
        return model(**resp.json())

    def _get_models(
        self, url: str, model: Type[BaseModel], limit: int = 500
    ) -> Iterator[BaseModel]:
        next_url = update_query(url, {"limit": limit})
        while next_url:
            resp = self._get(next_url)
            obj_list = model(**resp.json())
            for obj in obj_list.items:
                yield obj
            next_url = getattr(obj_list.links, "next", None)

    def _post(self, url: str, json_dict: dict) -> requests.Response:
        resp = self._session.post(url, json=json_dict)
        if resp.status_code >= 400:
            raise ValueError(resp.content)
        self._get.cache_clear()
        return resp

    def _delete(self, url: str) -> requests.Response:
        resp = self._session.delete(url)
        if resp.status_code >= 400:
            raise ValueError(resp.content)
        self._get.cache_clear()
        return resp

    def user(self) -> MyUser:
        """User object corresponding to session user

        Raises ``ValueError`` if user cannot be fetched.
        """
        return self.resolve(ME_URL.format(api_url=self.api_url), MyUser)

    def get_user(self, user_id: str) -> User:
        """User object corresponding to the id provided

        Raises ``ValueError`` if user cannot be fetched.
        """
        return self.resolve(
            USERS_URL.format(api_url=self.api_url, user_id=user_id), User
        )

    def my_projects(self, owner: bool = False) -> List[Project]:
        """All projects accessible to the user

        If ``owner = True``, only projects created by the user are returned

        Raises ``ValueError`` if projects cannot be fetched.
        """
        projects = self._get_models(self.user().links.projects, ProjectList)
        if owner:
            return [proj for proj in projects if proj.created_by == self.user().id]
        return list(projects)

    def get_project(self, project: str) -> Project:
        """Retrieve project based on identifier

        First this function tries to match ``project.id``.
        If no match is found, it tries to match ``project.name``.
        Finally, it tries to match ``project.title``.

        Raises ``ValueError`` if the title is non-unique or if no match is found.
        """
        my_projects = self.my_projects()
        for project_obj in my_projects:
            if project_obj.id == project:
                return project_obj
        for project_obj in my_projects:
            if project_obj.name == project:
                return project_obj
        titles = [project_obj.title for project_obj in my_projects]
        if titles.count(project) > 1:
            raise ValueError(f"duplicate projects found: {project}")
        for project_obj in my_projects:
            if project_obj.title == project:
                return project_obj
        raise ValueError(f"unable to find project: {project}")

    def get_block(
        self,
        project: Union[str, Project],
        block: str,
        kind: Optional[BlockKind] = None,
    ) -> Block:
        """Retrieve block based on identifier

        This function will only search within provided ``project``.
        If ``project`` is a string, it will look up with :meth:`get_project`.

        First this function tries to match ``block.id``.
        If no match is found, it tries to match ``block.name``.
        Finally, it tries to match ``block.title``.

        Raises ``ValueError`` if the title is non-unique or if no match is found.
        """
        logging.info("get_block")
        project_obj = (
            project if isinstance(project, Project) else self.get_project(project)
        )
        possible_blocks = []
        blocks_url = project_obj.links.blocks
        logging.debug(blocks_url)
        if kind:
            blocks_url = update_query(blocks_url, {"kind": kind})
        for block_obj in self._get_models(blocks_url, BlockList):
            if block_obj.id.block == block:
                logging.debug(f"found block by id {block}")
                return block_obj
            if block in (block_obj.name, block_obj.title):
                possible_blocks.append(block_obj)
        for block_obj in possible_blocks:
            if block_obj.name == block:
                return block_obj
        titles = [block_obj.title for block_obj in possible_blocks]
        if titles.count(block) > 1:
            raise ValueError(f"duplicate blocks found for {block}")
        for block_obj in possible_blocks:
            if block_obj.title == block:
                return block_obj
        raise ValueError(f"unable to find {block}")

    def get_version(self, block: Block, version: int) -> BlockVersion:
        """
        Fetches the a specific version of the Block supplied
        """
        if version > block.latest_version:
            raise ValueError(
                f"trying to fetch version {version} when latest version is {block.latest_version}"
            )
        version_obj = self._get_model(
            f"{block.links.self}/versions/{version}", BlockVersion
        )
        return version_obj

    def get_version_with_children(
        self, block: Block, version: int, fmt: BlockFormat
    ) -> ChildrenList:
        """
        Fetches the children of the Block supplied at the specified version
        """
        if version > block.latest_version:
            raise ValueError(
                f"trying to fetch version {version} when latest version is {block.latest_version}"
            )
        try:
            url = f"{block.links.self}/versions/{version}/children/?format={fmt}"
            return self._get_model(url, ChildrenList)
        except ValueError as err:
            logging.error(f"Failed to fetch children {url}")
            raise err

    def download_version(
        self,
        version: BlockVersion,
        content_format: Optional[BlockFormat] = None,
    ) -> bytes:
        """Download version content in specified format

        Raises ``ValueError`` if version cannot be downloaded.
        """
        url = version.links.download
        if content_format:
            url = update_query(url, query={"format": content_format})
        resp = self._get(url)
        return resp.content

    def resolve(self, url: str, model=None):
        """Resolve URL into :ref:`models-returned-from-the-api`

        All API URLs are valid as well as some URLs from the website.
        If ``model`` is provided, the allowed output will be constrained
        to that type.

        Raises ``ValueError`` if URL cannot be resolved into a known model
        or the provided ``model``.
        """
        parsed = urlparse(url)
        base_url = urlunparse((parsed.scheme, parsed.netloc, "", "", "", ""))
        if base_url not in [self.api_url, self.site_url] or not parsed.path:
            raise ValueError(f"invalid url {url}")
        path_parts = parsed.path.split("/")
        if (
            base_url == self.site_url
            and len(path_parts) > 3
            and path_parts[1].startswith("@")
            and path_parts[3].startswith("!")
        ):
            output = self.get_block(path_parts[2], path_parts[3][1:])
            if model and not isinstance(output, model):
                raise ValueError(f"unable to resolve url {url} to {model.__name__}")
            return output
        if base_url == self.api_url:
            models = [model] if model else API_RESPONSE_MODELS
            resp = self._get(url)
            for mod in models:
                try:
                    return mod(**resp.json())
                except ValidationError:
                    continue
        raise ValueError(
            f"unable to resolve url {url}" + (f" to {model.__name__}" if model else "")
        )

    def upload_file(self, filename: str, content_type: str) -> AnyHttpUrl:
        """Upload file to storage bucket

        Returns reference to the uploaded file.
        Raises ``ValueError`` if upload fails.
        """
        upload_resp = self._post(
            UPLOAD_URL.format(api_url=self.api_url),
            json_dict={"content_type": content_type, "size": path.getsize(filename)},
        )
        storage_url = upload_resp.content.decode("utf-8")
        with open(filename, "rb") as file:
            resp = requests.put(
                storage_url, file.read(), headers={"Content-Type": content_type}
            )
        if resp.status_code >= 400:
            raise ValueError(f"File upload failed: {resp.content}")
        return storage_url

    def upload_block(self, block: BlockPost, project: Project) -> Block:
        """Upload a block to specified project

        Raises ``ValueError`` if upload fails.
        """
        resp = self._post(project.links.blocks, block.dict(exclude_unset=True))
        return Block(**resp.json())

    def upload_version(
        self,
        version: Union[ArticleVersionPost, ContentVersionPost, ImageVersionPost],
        block: Block,
    ) -> BlockVersion:
        """Upload verson of a block

        Currently only Content, Article, and Image blocks are supported.

        Raises ``ValueError`` if upload fails.
        """
        resp = self._post(block.links.versions, version.dict(exclude_unset=True))
        return BlockVersion(**resp.json())

    def article_version_from_file(
        self, filename: str, article: Block, title: Optional[str] = None
    ) -> BlockVersion:
        """Upload new version of local file to curvenote Article

        Article version ``title`` will be set if it is specified.
        Currently only markdown files with ``.md`` extension are supported.

        Raises ``ValueError`` for other file types, if article block is not
        of kind Article, or for failures during upload.
        """
        if article.kind != BlockKind.article:
            raise ValueError(f"block is not an article {article.id.block}")
        if filename.endswith(".md"):
            content = Markdown(filename, self)
        else:
            raise ValueError("only markdown is supported")
        return content.upload(article, title)

    def article_from_file(
        self,
        filename: str,
        project: Project,
        title: Optional[str] = None,
        article_id: Optional[str] = None,
    ) -> BlockVersion:
        """Upload a single file to an Article in a curvenote Project

        First, this tries to find an existing Article, based on ``article_id``,
        which may be the unique Article id or name. If no ``article_id`` is
        provided, it tries to find an existing Article based on ``title`` (or
        ``filename`` if ``title`` is not provided). If no matching Article
        is found, a new Article is created with name based on ``title``
        (or ``filename`` if ``title`` is not provided).

        After finding or creating the Article, a new Version is created
        with the given ``title`` using :meth:`article_version_from_file`.

        Raises ``ValueError`` if there is a problem creating a new Article
        or a new Version of the Article.
        """
        title = title if title is not None else filename_to_title(filename)
        name = title_to_name(title)
        try:
            article_block = self.get_block(
                project, article_id or name, kind=BlockKind.article
            )
        except ValueError:
            article_block = self.upload_block(
                BlockPost(
                    name=name,
                    kind=BlockKind.article,
                    tags=[],
                ),
                project,
            )
        return self.article_version_from_file(filename, article_block, title)

    def notebook_from_file(
        self,
        filename: str,
        project: Project,
        title: Optional[str] = None,
    ) -> BlockVersion:
        """
        Upload a single ipynb notebook to a Curvenote Project

        Note: this currently uses an api endpoint that will not honor existing version information
        the endpoint will only
        """
        title = title if title is not None else filename_to_title(filename)
        notebook = Notebook(filename, self)

        try:
            block = notebook.upload(project, title)
        except ValueError as err:
            raise ValueError("could not upload notebook") from err

        return block

    def documents_from_folder(
        self, folder: str, project: Project, verbose: bool = True
    ) -> Project:
        """Upload contents of a local folder to curvenote Project

        This will attempt to import all .md or .ipynb files from within the folder.
        Calls :meth:`article_from_file` on all .md files
        Calls :meth:`notebook_from_file` on all .ipynb files
        Currently nested folders are not supported.

        Raises ``ValueError`` if there is nothing to upload or upload fails.
        """
        # cannot do multi extension globs so stagger import
        files = sorted(glob(f"{folder}/*.md"))
        if not files:
            if not glob(folder):
                raise ValueError("folder is empty or does not exist")
            logger.info("no markdown (.md) files found")
        else:
            for file in files:
                self.article_from_file(file, project)
                if verbose:
                    logger.info(f"Uploaded...{file[-25:].ljust(25)}", end="\r")

        notebook_files = sorted(glob(f"{folder}/*.ipynb"))
        if not notebook_files:
            if not glob(folder):
                raise ValueError("folder is empty or does not exist")
            logger.info("no notebooks (.ipynb) found")
        else:
            for file in notebook_files:
                self.notebook_from_file(file, project)
                if verbose:
                    logger.info(f"Uploaded...{file[-25:].ljust(25)}")
        if verbose:
            logger.info("\nDone!")
        return project

    def get_or_create_project(
        self,
        title: str,
        team: Optional[str] = None,
    ) -> Project:
        """Get curvenote Project based on title

        First, this tries to find an existing Project, based on ``title``.
        If Project cannot be found, a new Project is created.

        New projects are created under ``team``, if provided;
        otherwise they are created under the current user.
        For existing Projects, ``team`` is ignored.

        Raises ``ValueError`` if Project creation fails.
        """
        name = title_to_name(title)
        try:
            # TODO use /check/project/name service endpoint instead of relying on
            # exception
            project = self.get_project(name)
            logger.info("found project %s", name)
        except ValueError:
            logger.info("project %s not found, creating...", name)
            project_post = ProjectPost(
                team=team if team else self.user().username,
                name=name,
                title=title,
            )
            resp = self._post(
                PROJECTS_URL.format(api_url=self.api_url),
                project_post.dict(exclude_unset=True),
            )
            if resp.status_code > 201:
                raise ValueError(f"Error creating project status: {resp.status_code}")

            logger.info("project %s created", name)
            project = Project(**resp.json())
        return project

    def update_content(
        self,
        project: Project,
        content: str,
        content_format: BlockFormat,
        name: Optional[str] = None,
    ) -> BlockVersion:
        """Upload a new version of a content block

        If ``name`` is provided and a content block of that name is found
        with :meth:`get_block`, a new version of that block will be created.
        Otherwise a new block and new version will be created.
        """
        if name:
            try:
                block = self.get_block(project, name, kind=BlockKind.content)
            except ValueError:
                block = self.upload_block(
                    block=BlockPost(
                        kind=BlockKind.content,
                        name=name,
                    ),
                    project=project,
                )
        else:
            block = self.upload_block(
                block=BlockPost(
                    kind=BlockKind.content,
                ),
                project=project,
            )
        version = self.upload_version(
            version=ContentVersionPost(
                content=content,
                format=content_format,
            ),
            block=block,
        )
        return version

    def update_article(
        self,
        project: Project,
        child_versions: List[BlockVersion],
        title: str,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ) -> BlockVersion:
        """Upload a new version of an article block

        If an article block of ``title`` or ``name`` is found with
        :meth:`get_block`, a new version of that block will be created.
        Otherwise a new block and new version will be created.
        """
        if not name:
            name = title_to_name(title)
        if not tags:
            tags = []
        try:
            block = self.get_block(project, name, kind=BlockKind.article)
        except ValueError:
            block = self.upload_block(
                block=BlockPost(
                    kind=BlockKind.article,
                    name=name,
                    tags=tags,
                ),
                project=project,
            )
        version = self.upload_version(
            version=ArticleVersionPost(
                order=list(str(i) for i in range(len(child_versions))),
                children={
                    str(i): BlockChild(id=str(i), src=ver.id)
                    for i, ver in enumerate(child_versions)
                },
                title=title,
            ),
            block=block,
        )
        return version

    def update_navigation(
        self, project: Project, articles: List[Block]
    ) -> BlockVersion:
        """Update article order in project navigation

        Add the ``articles`` in the given order to the ``project`` navigation.
        This also deletes the existing draft, so the new version is visible.

        Raises ``ValueError`` if Blocks are invalid or update fails.
        """
        for article in articles:
            if article.kind != BlockKind.article:
                raise ValueError("only article blocks may be added to navigation")
            if article.id.project != project.id:
                raise ValueError(
                    "articles must be part of project to add to navigation"
                )
        nav_blocks = list(
            self._get_models(
                update_query(project.links.blocks, {"kind": "Navigation"}),
                BlockList,
            )
        )
        if len(nav_blocks) != 1:
            raise ValueError(
                f"Unexpected number of navigation blocks in project: {len(nav_blocks)}"
            )
        nav_block = nav_blocks[0]
        nav_version = NavigationVersionPost(
            items=[
                NavigationItem(
                    id=str(i),
                    title=article.title,
                    blockId=article.id,
                )
                for i, article in enumerate(articles)
            ]
        )
        resp = self._post(nav_block.links.versions, json_dict=nav_version.dict())
        if nav_block.links.default_draft:
            self._delete(nav_block.links.default_draft)
        return BlockVersion(**resp.json())



from tempfile import NamedTemporaryFile
from typing import Optional

import requests

from .models import BlockKind, BlockPost, BlockVersion, ImageVersionPost, Project
from .utils import github_to_raw, is_url

IMAGE_TAG = (
    '<img src="{src}" alt="{alt}" title="{title}" align="{align}" width="{width}%">'
)
BLOCK_SRC = "block:{project}/{block}/{version}"


class Image:
    def __init__(self, filename: str, session, content_type: Optional[str] = None):
        self.session = session
        if content_type:
            self.content_type = content_type
        elif filename.lower().endswith("png"):
            self.content_type = "image/png"
        elif filename.lower().endswith("gif"):
            self.content_type = "image/gif"
        elif filename.lower().endswith("jpg") or filename.lower().endswith("jpeg"):
            self.content_type = "image/jpg"
        else:
            raise ValueError(f"unsupported image type: {filename}")
        self.filename = filename

    def upload(self, project: Project, title: Optional[str] = None) -> BlockVersion:
        """Upload an image to given project

        If the image file is a URL, the image will be downloaded and copied
        to user uploads in curvenote.

        Raises ``ValueError`` if upload fails.
        """
        if is_url(self.filename):
            self.filename = github_to_raw(self.filename)
            resp = requests.get(self.filename)
            if resp.status_code >= 400:
                raise ValueError(f"Unable to find image {self.filename}")
            with NamedTemporaryFile() as file:
                file.write(resp.content)
                path = self.session.upload_file(file.name, self.content_type)
        else:
            path = self.session.upload_file(self.filename, self.content_type)
        block = self.session.upload_block(
            BlockPost(kind=BlockKind.image, tags=[]), project
        )
        version = self.session.upload_version(
            ImageVersionPost(upload_path=path, title=title), block
        )
        return version

    def create_tag(
        self,
        project: Project,
        title: Optional[str] = None,
        alt: Optional[str] = None,
        align: str = "center",
        width: int = 70,
    ) -> str:
        version = self.upload(project, title or alt)
        src = BLOCK_SRC.format(**version.id.dict())
        return IMAGE_TAG.format(
            title=title or "", alt=alt or "", src=src, align=align, width=width
        )



import os
import re
from typing import Dict
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from pydantic import AnyHttpUrl


def update_query(
    url: AnyHttpUrl,
    query: Dict[str, str],
) -> AnyHttpUrl:
    """Add query strings to URL"""
    parsed_url = urlparse(url)
    parsed_query = parse_qs(parsed_url.query)
    parsed_query.update(query)
    unparsed_url = urlunparse(
        tuple(parsed_url[:4] + (urlencode(parsed_query, doseq=True),) + parsed_url[5:])
    )
    return unparsed_url


def is_url(value: str) -> bool:
    """Returns True if value is URL"""
    return urlparse(value).scheme and urlparse(value).netloc


def github_to_raw(url: AnyHttpUrl) -> AnyHttpUrl:
    """Converts a github.com URL to raw.githubusercontent.com URL...

    Sketchy.
    """
    return url.replace("github.com", "raw.githubusercontent.com").replace("/blob/", "/")


def title_to_name(title: str) -> str:
    """Replace all non-alphanumeric characters with hyphens"""
    name = re.sub("[^0-9a-z]+", "-", title.lower())[:50]
    while name.startswith("-"):
        name = name[1:]
    while name.endswith("-"):
        name = name[:-1]
    return name


def filename_to_title(filename: str) -> str:
    """Remove path and extension from filename"""
    return filename.split(os.path.sep)[-1].rsplit(".", 1)[0]
